@@ -98,7 +98,6 @@ class ConjectureRunner(object):
         self.target_selector = TargetSelector(self.random)
 
         self.interesting_examples = {}
-        self.covering_examples = {}
 
         self.shrunk_examples = set()
 
@@ -422,6 +421,7 @@ class ConjectureRunner(object):
             choices = data.block_starts.get(n, [])
             if choices:
                 i = self.random.choice(choices)
+                assert i + n <= len(data.buffer)
                 return hbytes(data.buffer[i : i + n])
             else:
                 result = uniform(self.random, n)
@@ -481,11 +481,13 @@ class ConjectureRunner(object):
             if data.index + n > len(target_data[0].buffer):
                 result = uniform(self.random, n)
             else:
-                result = self.random.choice(bits)(data, n)
+                draw = self.random.choice(bits)
+                result = draw(data, n)
             p = prefix[0]
             if data.index < len(p):
                 start = p[data.index : data.index + n]
                 result = start + result[len(start) :]
+            assert len(result) == n
             return self.__zero_bound(data, result)
 
         return mutate_from
@@ -646,8 +648,6 @@ class ConjectureRunner(object):
                     result = uniform(self.random, n)
                 return self.__zero_bound(data, result)
 
-            targets_found = len(self.covering_examples)
-
             last_data = ConjectureData(
                 max_length=self.settings.buffer_size, draw_bytes=draw_bytes
             )
@@ -702,16 +702,12 @@ class ConjectureRunner(object):
             else:
                 origin = self.target_selector.select()
                 mutations += 1
-                targets_found = len(self.covering_examples)
                 data = ConjectureData(
                     draw_bytes=mutator(origin), max_length=self.settings.buffer_size
                 )
                 self.test_function(data)
                 data.freeze()
-                if (
-                    data.status > origin.status
-                    or len(self.covering_examples) > targets_found
-                ):
+                if data.status > origin.status:
                     mutations = 0
                 elif data.status < origin.status or mutations >= 10:
                     # Cap the variations of a single example and move on to
