@@ -17,23 +17,24 @@
 
 from __future__ import absolute_import, division, print_function
 
-import heapq
 from collections import defaultdict
-from enum import Enum
-from functools import total_ordering
 
 import attr
 
 from hypothesis.internal.compat import hbytes, hrange, int_from_bytes, int_to_bytes
-from hypothesis.internal.conjecture.data import Overrun, Status
+from hypothesis.internal.conjecture.data import ConjectureResult, Overrun, Status
 from hypothesis.internal.conjecture.floats import (
     DRAW_FLOAT_LABEL,
     float_to_lex,
     lex_to_float,
 )
+from hypothesis.internal.conjecture.junkdrawer import replace_all
 from hypothesis.internal.conjecture.shrinking import Float, Integer, Lexical, Ordering
 from hypothesis.internal.conjecture.shrinking.common import find_integer
 
+if False:
+    from typing import Dict  # noqa
+
 
 def sort_key(buffer):
     """Returns a sort key such that "simpler" buffers are smaller than
@@ -61,7 +62,11 @@ def sort_key(buffer):
     return (len(buffer), buffer)
 
 
-def defines_shrink_pass(generate_arguments):
+SHRINK_PASS_DEFINITIONS = {}  # type: Dict[str, ShrinkPassDefinition]
+
+
+@attr.s()
+class ShrinkPassDefinition(object):
     """A shrink pass bundles together a large number of local changes to
     the current shrink target.
 
@@ -76,35 +81,30 @@ def defines_shrink_pass(generate_arguments):
     changes.
     """
 
-    def accept(fn):
-        def run(self):
-            # We always maintain ``arguments`` as the current collection
-            # that would be calculated from self, and instead iterate over
-            # indices into it. This means we don't have to duplicate validation
-            # logic between it and the underlying shrink pass function.
-            #
-            # Note that in the event that the shrink target actually changes
-            # this can cause us to miss a few things that we might reasonably
-            # have run. e.g. if we successfully deleted an item then every
-            # index after that point now has an off by one error. This can
-            # cause redundant work but is not a correctness issue, because
-            # we guarantee that if the collection does not change then we will
-            # iterate over every argument, and this is satisfied if the shrink
-            # target never changes when running this pass.
-            arguments = list(generate_arguments(self))
-            indices = list(hrange(len(arguments)))
-            self.random.shuffle(indices)
-            while indices:
-                i = indices.pop()
-                if i >= len(arguments):
-                    continue
-                prev = self.shrink_target
-                args = tuple(arguments[i])
-                fn(self, *args)
-                if prev is not self.shrink_target:
-                    arguments = list(generate_arguments(self))
+    run_step = attr.ib()
+    generate_arguments = attr.ib()
 
-        run.__name__ = fn.__name__
+    @property
+    def name(self):
+        return self.run_step.__name__
+
+    def __attrs_post_init__(self):
+        assert self.name not in SHRINK_PASS_DEFINITIONS, self.name
+        SHRINK_PASS_DEFINITIONS[self.name] = self
+
+
+def defines_shrink_pass(generate_arguments):
+    """A convenient decorator for defining shrink passes."""
+
+    def accept(run_step):
+        definition = ShrinkPassDefinition(
+            generate_arguments=generate_arguments, run_step=run_step
+        )
+
+        def run(self):
+            return self.run_shrink_pass(definition.name)
+
+        run.__name__ = run_step.__name__
         run.is_shrink_pass = True
         return run
 
@@ -247,45 +247,6 @@ class Shrinker(object):
 
     """
 
-    def default_passes(self):
-        """Returns the list of shrink passes that are always safe to run,
-        in a good order to run them in.
-
-        These mostly have time
-        complexity that is at most O(n log(n)) in the size of the underlying
-        buffer. pass_to_descendant is an exception in that it technically
-        has worst case complexity O(n^2), but it is rare for it to hit
-        that case and generally runs very few operations at all.
-        """
-        return [
-            "alphabet_minimize",
-            "pass_to_descendant",
-            "zero_examples",
-            "adaptive_example_deletion",
-            "reorder_examples",
-            "minimize_floats",
-            "minimize_duplicated_blocks",
-            "minimize_individual_blocks",
-        ]
-
-    def emergency_passes(self):
-        """Returns the list of emergency shrink passes,
-        in a good order to run them in.
-
-        Emergency passes are ones that we hope don't do anything
-        very useful. The ideal scenario is that we run all of our
-        default passes to a fixed point, then we run all of the
-        emergency passes and they do nothing and we're finished.
-        This is because they're either a bit weird and designed to
-        handle some special case that doesn't come up very often,
-        or because they're very expensive, or both.
-        """
-        return [
-            block_program("-XX"),
-            block_program("XX"),
-            "example_deletion_with_block_lowering",
-        ]
-
     def derived_value(fn):
         """It's useful during shrinking to have access to derived values of
         the current shrink target.
@@ -314,20 +275,11 @@ class Shrinker(object):
         """
         self.__engine = engine
         self.__predicate = predicate
-        self.discarding_failed = False
         self.__shrinking_prefixes = set()
         self.__derived_values = {}
 
         self.initial_size = len(initial.buffer)
 
-        # We add a second level of caching local to the shrinker. This is a bit
-        # of a hack. Ideally we'd be able to rely on the engine's functionality
-        # for this. Doing it this way has two benefits: Firstly, the engine
-        # does not currently cache overruns (and probably shouldn't, but could
-        # recreate them on demand if necessary), and secondly Python dicts are
-        # much faster than our pure Python tree-based lookups.
-        self.__test_function_cache = {}
-
         # We keep track of the current best example on the shrink_target
         # attribute.
         self.shrink_target = None
@@ -336,91 +288,30 @@ class Shrinker(object):
 
         self.initial_calls = self.__engine.call_count
 
-        self.current_pass_depth = 0
         self.passes_by_name = {}
-        self.clear_passes()
-
-        for p in self.default_passes():
-            self.add_new_pass(p)
-
-        for p in self.emergency_passes():
-            self.add_new_pass(p, classification=PassClassification.AVOID)
-
-        self.add_new_pass(
-            "lower_common_block_offset", classification=PassClassification.SPECIAL
-        )
-
-    def clear_passes(self):
-        """Reset all passes on the shrinker, leaving it in a blank state.
-
-        This is mostly useful for testing.
-        """
-        # Note that we deliberately do not clear passes_by_name. This means
-        # that we can still look up and explicitly run the standard passes,
-        # they just won't be avaiable by default.
-
         self.passes = []
-        self.passes_awaiting_requeue = []
-        self.pass_queues = {c: [] for c in PassClassification}
 
-    def add_new_pass(self, run, classification=None):
+    def add_new_pass(self, run):
         """Creates a shrink pass corresponding to calling ``run(self)``"""
-        if classification is None:
-            classification = PassClassification.CANDIDATE
-        if isinstance(run, str):
-            run = getattr(Shrinker, run)
+
+        definition = SHRINK_PASS_DEFINITIONS[run]
+
         p = ShrinkPass(
-            pass_function=run, index=len(self.passes), classification=classification
+            run_with_arguments=definition.run_step,
+            generate_arguments=definition.generate_arguments,
+            shrinker=self,
+            index=len(self.passes),
         )
         self.passes.append(p)
-        self.passes_awaiting_requeue.append(p)
         self.passes_by_name[p.name] = p
         return p
 
     def shrink_pass(self, name):
         """Return the ShrinkPass object for the pass with the given name."""
-        if hasattr(Shrinker, name) and name not in self.passes_by_name:
-            self.add_new_pass(name, classification=PassClassification.SPECIAL)
+        if name not in self.passes_by_name:
+            self.add_new_pass(name)
         return self.passes_by_name[name]
 
-    def requeue_passes(self):
-        """Move all passes from passes_awaiting_requeue to their relevant
-        queues."""
-        while self.passes_awaiting_requeue:
-            p = self.passes_awaiting_requeue.pop()
-            heapq.heappush(self.pass_queues[p.classification], p)
-
-    def has_queued_passes(self, classification):
-        """Checks if any shrink passes are currently enqued under this
-        classification (note that there may be passes with this classification
-        currently awaiting requeue)."""
-        return len(self.pass_queues[classification]) > 0
-
-    def pop_queued_pass(self, classification):
-        """Pop and run a single queued pass with this classification."""
-        sp = heapq.heappop(self.pass_queues[classification])
-        self.passes_awaiting_requeue.append(sp)
-        self.run_shrink_pass(sp)
-
-    def run_queued_until_change(self, classification):
-        """Run passes with this classification until there are no more or one
-        of them succeeds in shrinking the target."""
-        initial = self.shrink_target
-        while self.has_queued_passes(classification) and self.shrink_target is initial:
-            self.pop_queued_pass(classification)
-        return self.shrink_target is not initial
-
-    def run_one_queued_pass(self, classification):
-        """Run a single queud pass with this classification (if there are
-        any)."""
-        if self.has_queued_passes(classification):
-            self.pop_queued_pass(classification)
-
-    def run_queued_passes(self, classification):
-        """Run all queued passes with this classification."""
-        while self.has_queued_passes(classification):
-            self.pop_queued_pass(classification)
-
     @property
     def calls(self):
         """Return the number of calls that have been made to the underlying
@@ -495,69 +386,19 @@ class Shrinker(object):
         this shrinker. This does not handle any requeing that is
         required.
         """
-        if isinstance(sp, str):
-            sp = self.shrink_pass(sp)
+        sp = self.shrink_pass(sp)
 
         self.debug("Shrink Pass %s" % (sp.name,))
-
-        initial_shrinks = self.shrinks
-        initial_calls = self.calls
-        size = len(self.shrink_target.buffer)
         try:
-            sp.pass_function(self)
-        finally:
-            calls = self.calls - initial_calls
-            shrinks = self.shrinks - initial_shrinks
-            deletions = size - len(self.shrink_target.buffer)
-
-            sp.calls += calls
-            sp.shrinks += shrinks
-            sp.deletions += deletions
             sp.runs += 1
+
+            steps = sp.generate_steps()
+            self.random.shuffle(steps)
+            for s in steps:
+                sp.run_step(s)
+        finally:
             self.debug("Shrink Pass %s completed." % (sp.name,))
 
-        # Complex state machine alert! A pass run can either succeed (we made
-        # at least one shrink) or fail (we didn't). This changes the pass's
-        # current classification according to the following possible
-        # transitions:
-        #
-        # CANDIDATE -------> HOPEFUL
-        #     |                 ^
-        #     |                 |
-        #     v                 v
-        #   AVOID ---------> DUBIOUS
-        #
-        # From best to worst we want to run HOPEFUL, CANDIDATE, DUBIOUS, AVOID.
-        # We will try any one of them if we have to but we want to prioritise.
-        #
-        # When a run succeeds, a pass will follow an arrow to a better class.
-        # When it fails, it will follow an arrow to a worse one.
-        # If no such arrow is available, it stays where it is.
-        #
-        # We also have the classification SPECIAL for passes that do not get
-        # run as part of the normal process.
-        previous = sp.classification
-
-        # If the pass didn't actually do anything we don't reclassify it. This
-        # is for things like remove_discarded which often are inapplicable.
-        if calls > 0 and sp.classification != PassClassification.SPECIAL:
-            if shrinks == 0:
-                if sp.successes > 0:
-                    sp.classification = PassClassification.DUBIOUS
-                else:
-                    sp.classification = PassClassification.AVOID
-            else:
-                sp.successes += 1
-                if sp.classification == PassClassification.AVOID:
-                    sp.classification = PassClassification.DUBIOUS
-                else:
-                    sp.classification = PassClassification.HOPEFUL
-            if previous != sp.classification:
-                self.debug(
-                    "Reclassified %s from %s to %s"
-                    % (sp.name, previous.name, sp.classification.name)
-                )
-
     def shrink(self):
         """Run the full set of shrinks and update shrink_target.
 
@@ -647,69 +488,78 @@ class Shrinker(object):
         This method iterates to a fixed point and so is idempontent - calling
         it twice will have exactly the same effect as calling it once.
         """
-        while self.single_greedy_shrink_iteration():
-            self.run_shrink_pass("lower_common_block_offset")
 
-    def single_greedy_shrink_iteration(self):
-        """Performs a single run through each greedy shrink pass, but does not
-        loop to achieve a fixed point."""
-        initial = self.shrink_target
+        # "coarse" passes are ones which either make large scale modifications
+        # to the test case (alphabet_minimize) or delete data from it (the
+        # rest). After these have reached a fixed point the test case should
+        # be reasonably small and well normalized.
+        coarse = [
+            "alphabet_minimize",
+            "pass_to_descendant",
+            "zero_examples",
+            "adaptive_example_deletion",
+        ]
+        self.fixate_shrink_passes(coarse)
 
-        # What follows is a slightly delicate dance. What we want to do is try
-        # to ensure that:
-        #
-        # 1. If it is possible for us to be deleting data, we should be.
-        # 2. We do not end up repeating a lot of passes uselessly.
-        # 3. We do not want to run expensive or useless passes if we can
-        #    possibly avoid doing so.
+        # "fine" passes are ones that make lots of fine grained changes
+        # to the shrink target. Typically we might expect these to make many
+        # small changes, with some changes unlocking others. Additionally,
+        # if there's more data left to delete then often running these is
+        # wasteful. As a result we only start running them after we've hit
+        # a fixed point for the coarse passes at least once.
+        fine = [
+            "reorder_examples",
+            "minimize_floats",
+            "minimize_duplicated_blocks",
+            "minimize_individual_blocks",
+        ]
 
-        self.requeue_passes()
+        self.fixate_shrink_passes(coarse + fine)
 
-        self.run_shrink_pass("remove_discarded")
+        # "emergency" shrink passes are ones that handle cases that we
+        # can't currently handle more elegantly - either they're slightly
+        # weird hacks that happen to work or they're expensive passes to
+        # run. Generally we hope that the emergency passes don't do anything
+        # at all.
+        emergency = [
+            block_program("-XX"),
+            block_program("XX"),
+            "example_deletion_with_block_lowering",
+        ]
 
-        # First run the entire set of solid passes (ones that have previously
-        # made changes). It's important that we run all of them, not just one,
-        # as typically each pass may unlock others.
-        self.run_queued_passes(PassClassification.HOPEFUL)
+        self.fixate_shrink_passes(coarse + fine + emergency)
 
-        # While our solid passes are successfully shrinking the buffer, we can
-        # just keep doing that (note that this is a stronger condition than
-        # just making shrinks - it's a much better sense of progress. We can
-        # make only O(n) length reductions but we can make exponentially many
-        # shrinks).
-        if len(self.buffer) < len(initial.buffer):
-            return True
+    def fixate_shrink_passes(self, passes):
+        """Run steps from each pass in ``passes`` until the current shrink target
+        is a fixed point of all of them."""
+        passes = list(map(self.shrink_pass, passes))
 
-        # If we're stuck on length reductions then we pull in one candiate pass
-        # (if there are any).
-        # This should hopefully help us unlock any local minima that were
-        # starting to reduce the utility of the previous solid passes.
-        self.run_one_queued_pass(PassClassification.CANDIDATE)
+        initial = None
+        while initial is not self.shrink_target:
+            initial = self.shrink_target
+            for sp in passes:
+                if sp.arguments:
+                    sp.runs += 1
 
-        # We've pulled in a new candidate pass (or have no candidate passes
-        # left) and are making shrinks with the solid passes, so lets just
-        # keep on doing that.
-        if self.shrink_target is not initial:
-            return True
+            passes_with_steps = [
+                (sp, step) for sp in passes for step in sp.generate_steps()
+            ]
+            self.random.shuffle(passes_with_steps)
 
-        # We're a bit stuck, so it's time to try some new passes.
-        for classification in [
-            # First we try rerunning every pass we've previously seen succeed.
-            PassClassification.DUBIOUS,
-            # If that didn't work, we pull in some new candidate passes.
-            PassClassification.CANDIDATE,
-            # If that still didn't work, we now pull out all the stops and
-            # bring in the desperation passes. These are either passes that
-            # started as CANDIDATE but we have never seen work, or ones that
-            # are so expensive that they begin life as AVOID.
-            PassClassification.AVOID,
-        ]:
-            if self.run_queued_until_change(classification):
-                return True
+            # We run remove_discarded after every step to do cleanup
+            # keeping track of whether that actually works. Either there is
+            # no discarded data and it is basically free, or it reliably works
+            # and deletes data, or it doesn't work. In that latter case we turn
+            # it off for the rest of this loop through the passes, but will
+            # try again once all of the passes have been run.
+            can_discard = self.remove_discarded()
 
-        assert self.shrink_target is initial
-
-        return False
+            for sp, step in passes_with_steps:
+                sp.run_step(step)
+                if can_discard:
+                    can_discard &= self.remove_discarded()
+        for sp in passes:
+            sp.fixed_point_at = self.shrink_target
 
     @property
     def buffer(self):
@@ -723,8 +573,10 @@ class Shrinker(object):
     def examples(self):
         return self.shrink_target.examples
 
-    def all_block_bounds(self):
-        return self.shrink_target.all_block_bounds()
+    def all_block_bounds(self, target=None):
+        if target is None:
+            target = self.shrink_target
+        return [(b.start, b.end) for b in target.blocks]
 
     @derived_value
     def examples_by_label(self):
@@ -806,7 +658,6 @@ class Shrinker(object):
             i, t.buffer[: t.blocks[i].start] in self.__shrinking_prefixes
         )
 
-    @defines_shrink_pass(lambda self: [()])
     def lower_common_block_offset(self):
         """Sometimes we find ourselves in a situation where changes to one part
         of the byte stream unlock changes to other parts. Sometimes this is
@@ -845,7 +696,7 @@ class Shrinker(object):
 
         current = self.shrink_target
 
-        blocked = [current.buffer[u:v] for u, v in current.all_block_bounds()]
+        blocked = [current.buffer[u:v] for u, v in self.all_block_bounds(current)]
 
         changed = [
             i
@@ -869,9 +720,8 @@ class Shrinker(object):
                 new_blocks[i] = int_to_bytes(v + o, len(blocked[i]))
             return self.incorporate_new_buffer(hbytes().join(new_blocks))
 
-        new_offset = Integer.shrink(offset, reoffset, random=self.random)
-        if new_offset == offset:
-            self.clear_change_tracking()
+        Integer.shrink(offset, reoffset, random=self.random)
+        self.clear_change_tracking()
 
     def mark_shrinking(self, blocks):
         """Mark each of these blocks as a shrinking block: That is, lowering
@@ -891,7 +741,7 @@ class Shrinker(object):
         self.__changed_blocks.add(i)
 
     def update_shrink_target(self, new_target):
-        assert new_target.frozen
+        assert isinstance(new_target, ConjectureResult)
         if self.shrink_target is not None:
             current = self.shrink_target.buffer
             new = new_target.buffer
@@ -899,7 +749,7 @@ class Shrinker(object):
             self.shrinks += 1
             if (
                 len(new_target.blocks) != len(self.shrink_target.blocks)
-                or new_target.all_block_bounds() != self.all_block_bounds()
+                or self.all_block_bounds(new_target) != self.all_block_bounds()
             ):
                 self.clear_change_tracking()
             else:
@@ -937,12 +787,16 @@ class Shrinker(object):
             n = min(self.blocks[block].length, len(b))
             initial_attempt[v - n : v] = b[-n:]
 
+        if not blocks:
+            return False
+
         start = self.shrink_target.blocks[blocks[0]].start
         end = self.shrink_target.blocks[blocks[-1]].end
 
         initial_data = self.cached_test_function(initial_attempt)
 
         if initial_data.status == Status.INTERESTING:
+            self.lower_common_block_offset()
             return initial_data is self.shrink_target
 
         # If this produced something completely invalid we ditch it
@@ -1022,11 +876,10 @@ class Shrinker(object):
                 return True
         return False
 
-    @defines_shrink_pass(lambda self: [()] if self.shrink_target.has_discards else [])
     def remove_discarded(self):
         """Try removing all bytes marked as discarded.
 
-        This pass is primarily to deal with data that has been ignored while
+        This is primarily to deal with data that has been ignored while
         doing rejection sampling - e.g. as a result of an integer range, or a
         filtered strategy.
 
@@ -1035,22 +888,35 @@ class Shrinker(object):
         each interval individually. The common case is that all data drawn and
         rejected can just be thrown away immediately in one block, so this pass
         will be much faster than trying each one individually when it works.
+
+        returns False if there is discarded data and removing it does not work,
+        otherwise returns True.
         """
         while self.shrink_target.has_discards:
             discarded = []
 
             for ex in self.shrink_target.examples:
-                if ex.discarded and (not discarded or ex.start >= discarded[-1][-1]):
+                if (
+                    ex.length > 0
+                    and ex.discarded
+                    and (not discarded or ex.start >= discarded[-1][-1])
+                ):
                     discarded.append((ex.start, ex.end))
 
-            assert discarded
+            # This can happen if we have discards but they are all of
+            # zero length. This shouldn't happen very often so it's
+            # faster to check for it here than at the point of example
+            # generation.
+            if not discarded:
+                break
 
             attempt = bytearray(self.shrink_target.buffer)
             for u, v in reversed(discarded):
                 del attempt[u:v]
 
             if not self.incorporate_new_buffer(attempt):
-                break
+                return False
+        return True
 
     @derived_value
     def endpoints_by_depth(self):
@@ -1109,9 +975,8 @@ class Shrinker(object):
         if to_right > 0:
             find_integer(lambda n: delete_region(j - n, j + to_right))
 
-    @defines_shrink_pass(lambda self: [(e,) for e in self.examples if not e.trivial])
-    def zero_examples(self, ex):
-        """Attempt to replace each example with a minimal version of itself."""
+    def try_zero_example(self, i):
+        ex = self.examples[i]
         u = ex.start
         v = ex.end
         attempt = self.cached_test_function(
@@ -1119,19 +984,66 @@ class Shrinker(object):
         )
 
         if attempt is Overrun:
-            return
+            return False
 
         in_replacement = attempt.examples[ex.index]
         used = in_replacement.length
 
-        if (
-            not self.__predicate(attempt)
-            and in_replacement.end < len(attempt.buffer)
-            and used < ex.length
-        ):
-            self.incorporate_new_buffer(
-                self.buffer[:u] + hbytes(used) + self.buffer[v:]
-            )
+        if attempt is not self.shrink_target:
+            if in_replacement.end < len(attempt.buffer) and used < ex.length:
+                self.incorporate_new_buffer(
+                    self.buffer[:u] + hbytes(used) + self.buffer[v:]
+                )
+        return self.examples[i].trivial
+
+    @defines_shrink_pass(lambda self: [(e.index,) for e in self.examples])
+    def zero_examples(self, ex_index):
+        """Attempt to replace each example with a minimal version of itself."""
+        # If the example is already trivial, assume there's nothing to do here.
+        # We could attempt to use it as an adaptive replacement for other
+        # similar examples, but that seems to be ineffective, resulting mostly
+        # in redundant work rather than helping.
+        if self.examples[ex_index].trivial:
+            return
+
+        if not self.try_zero_example(ex_index):
+            return
+
+        ex = self.examples[ex_index]
+
+        original = self.shrink_target
+        group = self.examples_by_label[ex.label]
+        i = group.index(ex)
+        replacement = self.buffer[ex.start : ex.end]
+
+        # We first expand to cover the trivial region surrounding this group.
+        # This avoids a situation where the adaptive phase "succeeds" a lot by
+        # virtue of not doing anything and then goes into a galloping phase
+        # where it does a bunch of useless work.
+        def all_trivial(a, b):
+            if a < 0 or b > len(group):
+                return False
+            return all(e.trivial for e in group[a:b])
+
+        start, end = expand_region(all_trivial, i, i + 1)
+
+        # If we've got multiple trivial examples of different lengths then
+        # this isn't going to work as a replacement for all of them and so we
+        # skip out early.
+        if any(e.length != len(replacement) for e in group[start:end]):
+            return
+
+        def can_zero(a, b):
+            if a < 0 or b > len(group):
+                return False
+            regions = []
+            for e in group[a:b]:
+                t = (e.start, e.end, replacement)
+                if not regions or t[0] >= regions[-1][1]:
+                    regions.append(t)
+            return self.consider_new_buffer(replace_all(original.buffer, regions))
+
+        expand_region(can_zero, start, end)
 
     def duplicated_block_suffixes(self):
         """Returns a list of blocks grouped by their non-zero suffix,
@@ -1280,12 +1192,19 @@ class Shrinker(object):
         del buf[ex.start : ex.end]
         self.incorporate_new_buffer(buf)
 
+    @derived_value
+    def groups_for_reordering(self):
+        result = defaultdict(list)
+        for e in self.examples:
+            result[(e.parent, e.label)].append(e)
+        groups = [v for v in result.values() if len(v) > 1]
+        groups.sort(key=lambda ls: ls[0].index)
+        return groups
+
     @defines_shrink_pass(
-        lambda self: [
-            (e,) for e in self.examples if not e.trivial and len(e.children) > 1
-        ]
+        lambda self: [(i,) for i in hrange(len(self.groups_for_reordering))]
     )
-    def reorder_examples(self, ex):
+    def reorder_examples(self, i):
         """This pass allows us to reorder the children of each example.
 
         For example, consider the following:
@@ -1303,13 +1222,16 @@ class Shrinker(object):
         ``x=""``, ``y="0"``, or the other way around. With reordering it will
         reliably fail with ``x=""``, ``y="0"``.
         """
+        group = self.groups_for_reordering[i]
         st = self.shrink_target
-        pieces = [st.buffer[c.start : c.end] for c in ex.children]
-        prefix = st.buffer[: ex.start]
-        suffix = st.buffer[ex.end :]
+        pieces = [st.buffer[ex.start : ex.end] for ex in group]
+        endpoints = [(ex.start, ex.end) for ex in group]
+
         Ordering.shrink(
             pieces,
-            lambda ls: self.incorporate_new_buffer(prefix + hbytes().join(ls) + suffix),
+            lambda ls: self.consider_new_buffer(
+                replace_all(st.buffer, [(u, v, r) for (u, v), r in zip(endpoints, ls)])
+            ),
             random=self.random,
         )
 
@@ -1387,6 +1309,46 @@ class Shrinker(object):
             else:
                 lo = mid
 
+    def run_block_program(self, i, description, original, repeats=1):
+        """Block programs are a mini-DSL for block rewriting, defined as a sequence
+        of commands that can be run at some index into the blocks
+
+        Commands are:
+
+            * "-", subtract one from this block.
+            * "X", delete this block
+
+        If a command does not apply (currently only because it's - on a zero
+        block) the block will be silently skipped over.
+
+        This method runs the block program in ``description`` at block index
+        ``i`` on the ConjectureData ``original``. If ``repeats > 1`` then it
+        will attempt to approximate the results of running it that many times.
+
+        Returns True if this successfully changes the underlying shrink target,
+        else False.
+        """
+        if i + len(description) > len(original.blocks) or i < 0:
+            return False
+        attempt = bytearray(original.buffer)
+        for _ in hrange(repeats):
+            for k, d in reversed(list(enumerate(description))):
+                j = i + k
+                u, v = original.blocks[j].bounds
+                if v > len(attempt):
+                    return False
+                if d == "-":
+                    value = int_from_bytes(attempt[u:v])
+                    if value == 0:
+                        return False
+                    else:
+                        attempt[u:v] = int_to_bytes(value - 1, v - u)
+                elif d == "X":
+                    del attempt[u:v]
+                else:  # pragma: no cover
+                    assert False, "Unrecognised command %r" % (d,)
+        return self.incorporate_new_buffer(attempt)
+
 
 def block_program(description):
     """Mini-DSL for block rewriting. A sequence of commands that will be run
@@ -1402,83 +1364,105 @@ def block_program(description):
     block) the block will be silently skipped over. As a side effect of
     running a block program its score will be updated.
     """
+    name = "block_program(%r)" % (description,)
 
-    def generate_arguments(self):
-        return [(i,) for i in hrange(len(self.blocks))]
-
-    def run(self, i):
+    if name not in SHRINK_PASS_DEFINITIONS:
+        """Defines a shrink pass that runs the block program ``description``
+        at every block index."""
         n = len(description)
-        if i + n > len(self.shrink_target.blocks):
-            return
-        attempt = bytearray(self.shrink_target.buffer)
-        for k, d in reversed(list(enumerate(description))):
-            j = i + k
-            block = self.blocks[j]
-            u, v = block.bounds
-            if d == "-":
-                value = int_from_bytes(attempt[u:v])
-                if value == 0:
-                    return
-                else:
-                    attempt[u:v] = int_to_bytes(value - 1, block.length)
-            elif d == "X":
-                del attempt[u:v]
-            else:  # pragma: no cover
-                assert False, "Unrecognised command %r" % (d,)
-        self.incorporate_new_buffer(attempt)
 
-    run.__name__ = "block_program(%r)" % (description,)
+        def generate_arguments(self):
+            return [(i,) for i in hrange(len(self.blocks) - n)]
 
-    return defines_shrink_pass(generate_arguments)(run)
+        def run(self, i):
+            """Adaptively attempt to run the block program at the current
+            index. If this successfully applies the block program ``k`` times
+            then this runs in ``O(log(k))`` test function calls."""
+            assert i + n < len(self.shrink_target.blocks)
+
+            # First, run the block program at the chosen index. If this fails,
+            # don't do any extra work, so that failure is as cheap as possible.
+            if not self.run_block_program(i, description, original=self.shrink_target):
+                return
+
+            # Because we run in a random order we will often find ourselves in the middle
+            # of a region where we could run the block program. We thus start by moving
+            # left to the beginning of that region if possible in order to to start from
+            # the beginning of that region.
+            def offset_left(k):
+                return i - k * n
+
+            i = offset_left(
+                find_integer(
+                    lambda k: self.run_block_program(
+                        offset_left(k), description, original=self.shrink_target
+                    )
+                )
+            )
+
+            original = self.shrink_target
+
+            # Now try to run the block program multiple times here.
+            find_integer(
+                lambda k: self.run_block_program(
+                    i, description, original=original, repeats=k
+                )
+            )
+
+        run.__name__ = name
+
+        defines_shrink_pass(generate_arguments)(run)
+        assert name in SHRINK_PASS_DEFINITIONS
+    return name
 
 
-class PassClassification(Enum):
-    CANDIDATE = 0
-    HOPEFUL = 1
-    DUBIOUS = 2
-    AVOID = 3
-    SPECIAL = 4
-
-
-@total_ordering
 @attr.s(slots=True, cmp=False)
 class ShrinkPass(object):
-    pass_function = attr.ib()
+    run_with_arguments = attr.ib()
+    generate_arguments = attr.ib()
     index = attr.ib()
+    shrinker = attr.ib()
 
-    classification = attr.ib(default=PassClassification.CANDIDATE)
+    __arguments = attr.ib(default=None, init=False)
+    __target_at_argument_calculation = attr.ib(default=None, init=False)
 
+    fixed_point_at = attr.ib(default=None)
     successes = attr.ib(default=0)
     runs = attr.ib(default=0)
     calls = attr.ib(default=0)
     shrinks = attr.ib(default=0)
     deletions = attr.ib(default=0)
 
-    def __attrs_post_init__(self):
-        assert getattr(
-            self.pass_function, "is_shrink_pass", False
-        ), "%s is not a shrink pass" % (self.pass_function.__name__,)
-
     @property
-    def failures(self):
-        return self.runs - self.successes
+    def arguments(self):
+        if self.__target_at_argument_calculation is not self.shrinker.shrink_target:
+            self.__arguments = self.generate_arguments(self.shrinker)
+            self.__target_at_argument_calculation = self.shrinker.shrink_target
+        return self.__arguments
+
+    def generate_steps(self):
+        if self.fixed_point_at is self.shrinker.shrink_target:
+            return []
+        return list(hrange(len(self.arguments)))
+
+    def run_step(self, i):
+        if i >= len(self.arguments):
+            return
+        args = self.arguments[i]
+        self.shrinker.debug("%s(%s)" % (self.name, ", ".join(map(repr, args))))
+        initial_shrinks = self.shrinker.shrinks
+        initial_calls = self.shrinker.calls
+        size = len(self.shrinker.shrink_target.buffer)
+        try:
+            self.run_with_arguments(self.shrinker, *args)
+        finally:
+            self.calls += self.shrinker.calls - initial_calls
+            self.shrinks += self.shrinker.shrinks - initial_shrinks
+            self.deletions += size - len(self.shrinker.shrink_target.buffer)
 
     @property
     def name(self):
-        return self.pass_function.__name__
-
-    def __eq__(self, other):
-        return self.index == other.index
-
-    def __hash__(self):
-        return hash(self.index)
-
-    def __lt__(self, other):
-        return self.key() < other.key()
-
-    def key(self):
-        # Smaller is better.
-        return (self.runs, self.failures, self.calls, self.index)
+        return self.run_with_arguments.__name__
 
 
 def non_zero_suffix(b):
@@ -1488,3 +1472,12 @@ def non_zero_suffix(b):
     while i < len(b) and b[i] == 0:
         i += 1
     return b[i:]
+
+
+def expand_region(f, a, b):
+    """Attempts to find u, v with u <= a, v >= b such that f(u, v) is true.
+    Assumes that f(a, b) is already true.
+    """
+    b += find_integer(lambda k: f(a, b + k))
+    a -= find_integer(lambda k: f(a - k, b))
+    return (a, b)
